/*
 * Copyright (c) 2008, AIST.
 * All rights reserved. This program is made available under the terms of the
 * Eclipse Public License v1.0 which accompanies this distribution, and is
 * available at http://www.eclipse.org/legal/epl-v10.html
 * Contributors:
 * National Institute of Advanced Industrial Science and Technology (AIST)
 */
/*
 * socket.c
 *
 *      Copyright(c) Isao Hara, 2006
 *
 * $Revision: 1.3 $
 * $Date: 2008/05/29 04:54:56 $
 * $Id: socket.c,v 1.3 2008/05/29 04:54:56 yoshi Exp $
 *
 */

#include <math.h>
#include <RtORB/corba.h>
#include <RtORB/sockport.h>

#include <kernel_cfg.h>
#include <tinet/TINETWrapper.h>

struct sockport_profile  SockProfile[FD_SETSIZE];
static  fd_set main_socket_bits;
#ifdef USE_THREAD
static ID SOCK_MUTEX = ORB_SOCK_SEM1;
#define SOCKET_LOCK()	wai_sem( SOCK_MUTEX )
#define SOCKET_UNLOCK()	sig_sem( SOCK_MUTEX )
#else
#define SOCKET_LOCK() 
#define SOCKET_UNLOCK() 
#endif

/*
 *   Server Side
 */
int make_server_socket_port(int port_no) {
  int server_sock;
  struct sockaddr_in *server_addr;
  int reuse = 1;

  server_addr=(struct sockaddr_in *)RtORB_alloc(sizeof(struct sockaddr_in),
		  "make_server_socket_port");
  memset(server_addr, 0, sizeof (*server_addr));
  server_addr->sin_family      = AF_INET;
  server_addr->sin_port        = htons(port_no);
  server_addr->sin_addr.s_addr = htonl(INADDR_ANY);

  server_sock = socket (AF_INET, SOCK_STREAM, 0);

  if (server_sock < 0) {
	  ORB_ERROR_OUT1("Fail to create a socket\n");
	  free(server_addr);
	  return -1;
  }

  setsockopt(server_sock, SOL_SOCKET, SO_REUSEADDR, (char *)&reuse, sizeof(reuse));

  if (bind(server_sock, (struct sockaddr *)server_addr, sizeof(*server_addr)) < 0){
	ORB_ERROR_OUT1("Fail to bind a socket\n");
    close(server_sock);
    return -1;
  }

  free(server_addr);

  if (listen (server_sock, 2) < 0 ){
	  ORB_ERROR_OUT1("listen failed\n");
      close(server_sock);
      return -1;
  }
  return server_sock;
}

long accept_connection(long port_no) {
  struct sockaddr_in  client_addr;
  int   client_len, snew, opt_len, stat;
  long  opt;

  client_len = sizeof(struct sockaddr_in);
  snew = accept(port_no, (struct sockaddr *)&client_addr, &client_len);
 
  if (snew>0) {
    stat=getsockopt(snew, 6, SO_REUSEADDR, &opt, &opt_len);
    opt = 1;
    stat=setsockopt(snew, 6, SO_REUSEADDR, &opt, sizeof(opt));
    stat=getsockopt(snew, 6, SO_REUSEADDR, &opt, &opt_len);
    }
  return(snew);
}

int copy_fds_set(fd_set *target, fd_set *src){
  int i,j,k;
  int	 nfds = 1;

  for(k= FD_SETSIZE / NFDBITS ; k >0 ;k--){
#if CYGWIN
    target->fds_bits[k-1] |= src->fds_bits[k-1];
    if (target->fds_bits[k-1]) break;
#else
    target->__fds_bits[k-1] |= src->__fds_bits[k-1];
    if (target->__fds_bits[k-1]) break;
#endif
  }

  if (k) {
#if CYGWIN
    j=target->fds_bits[k-1];
#else
    j=target->__fds_bits[k-1];
#endif
    for (i=0; i < NFDBITS; i++, j = j >> 1 ) { if (j & 1) nfds=i+1; }
    nfds += NFDBITS * (k-1);
  }
  for(; k >0 ;k--){
#if CYGWIN
    target->fds_bits[k-1] |= src->fds_bits[k-1];
#else
    target->__fds_bits[k-1] |= src->__fds_bits[k-1];
#endif
  }

  return nfds;
}

int select_server_sockets(fd_set *sockbits, int server_port, 
	struct timeval time_out,
	void (*connection_request)(),
	int  (*command_request)())
{
  int	 nfds;
  fd_set socks;
  int	 stat, i = 0, newsock;

  FD_ZERO(&socks);
  nfds = copy_fds_set(&socks, sockbits);

  stat=select(nfds, &socks, 0, 0 /*&exceptions*/ , &time_out);
  if (stat>0) {
    if (FD_ISSET(server_port, &socks)) { 
		ORB_ERROR_OUT1("New Connection\n");
      newsock= accept_connection(server_port); /* new connection */
      if (newsock>0) FD_SET(newsock, sockbits);
      FD_CLR(server_port, &socks);
      if (connection_request) (*connection_request)(newsock);
      stat--;
    } else {
      i=0;
      while (stat > 0 && i < nfds) {
        if (FD_ISSET(i, &socks)) {
		ORB_ERROR_OUT2("Request!! (%d)\n", i);
          if((*command_request)(i) == -1) FD_CLR(i, sockbits);
          FD_CLR(i, &socks);
          stat--;
        }
      i++; 
      }
    }
  }
  return i; 
}

void server_socket_loop(int server_port, float time_out_float, void (*idle)(),
  void (*connection_request)(), 
  int (*command_request)())
{
  fd_set sockbits;
  struct timeval time_out;
  struct timezone tz;
  struct timeval time1,time2;
  long time_cost;

  FD_ZERO(&sockbits);
  FD_SET(server_port, &sockbits);

  time_out.tv_sec = (unsigned long)(time_out_float / 1000.0);
  time_out.tv_usec = (time_out_float - time_out.tv_sec * 1000) * 1000 ;

//  fprintf(stderr, "--time_out %d, %d\n", time_out.tv_sec, time_out.tv_usec);
  while (1) {
//      gettimeofday(&time1, &tz); 
      select_server_sockets(&sockbits, server_port, time_out,
		connection_request, command_request);

      if (idle) (*idle)(); 
//      gettimeofday(&time2, &tz); 

#if 1
      time_cost = (time2.tv_sec - time1.tv_sec) * 1000000 + time2.tv_usec
	      - time1.tv_usec - time_out_float * 1000 ;

//      if (time_cost > 0) usleep(time_cost);
//      else usleep(1);
#endif
  //    usleep(1);
  }
}

/**
 *
 *   find a hostname and a port number from a socket descriptor
 *
 */

short get_socket_port(int sock){
  int len;
  struct sockaddr_in addr;

  len = sizeof(struct sockaddr_in);

  if (getsockname(sock, (struct sockaddr *)&addr, &len) < 0) return -1;
  return  ntohs(addr.sin_port);
}

char *get_ip_address(int sock){

	struct sockaddr_in addr;
	char ip_addr[32];

	memset( ip_addr, 0, sizeof( ip_addr ) );
	memset( &addr, 0, sizeof( addr ) );
	if ( getsockname( sock, (struct sockaddr *)&addr, sizeof( addr ) ) == 0 )
	{
		t_vsnprintf_4(	ip_addr,
						sizeof( ip_addr ),
						"%d.%d.%d.%d",
						(int)addr.sin_addr.S_un.S_un_b.s_b1,
						(int)addr.sin_addr.S_un.S_un_b.s_b2,
						(int)addr.sin_addr.S_un.S_un_b.s_b3,
						(int)addr.sin_addr.S_un.S_un_b.s_b4	);

#if DEBUG1
		ORB_ERROR_OUT2("Host = %s \n", ip_addr);
#endif
		if ( strcmp( ip_addr, "127.0.0.1" ) && strcmp( ip_addr, "0.0.0.0" ) ) 
		{
			return RtORB_strdup(ip_addr, "get_ip_address");
		}
	}
	return (char *)NULL;
}

/*
 *
 * Client side
 *
 **/
int make_client_socket_port(char *hostname, int port)
{
	struct sockaddr_in sock_addr;
	int fd;
	unsigned int ip_addr = 0;
	char temp[16];
	int length;
	int index;
	int dot;
	int cp_index;
	int ip[4];

	if ( hostname == NULL )
	{
		return (-1);
	}

	length = strlen( hostname );
	if ( length >= 7 && length <= 15 )
	{
		memset( temp, '.', sizeof( temp ) );
		memcpy( temp, hostname, length );

		dot = 0;
		cp_index = 0;
		for ( index = 0; index <= length; index++ )
		{
			if ( temp[ index ] == '.')
			{
				temp[ index ] = '\0';
				ip[ dot ] = atoi( &temp[cp_index] );

				cp_index = index + 1;
				dot++;
			}
			else if ( temp[ index ] < '0' || temp[ index ] > '9' )
			{
				break;
			}
		}

		if ( dot == 4 )
		{
			ip_addr = ( ip[0] * 0x1000000 ) + ( ip[1] * 0x10000 ) + ( ip[2] * 0x100 ) + ip[3];
		}
	}

	if ( ip_addr == 0 )
	{
		ORB_ERROR_OUT2( "host %s not valid\r\n", hostname);
		return (-1);
	}

	sock_addr.sin_addr.s_addr = htonl(ip_addr);
	sock_addr.sin_family = AF_INET;
	sock_addr.sin_port = htons(port);

	/* create socket */
	if ((fd = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
		 ORB_ERROR_OUT1("socket");
		 return(-1);
	}

	if(connect(fd, (struct sockaddr *)&sock_addr, sizeof(sock_addr)) < 0){
		ORB_ERROR_OUT1("connect");
		return(-1);
	}
	
	return fd;
}

#if 0 /* no use */
int select_client_socket( int client_port, int timeout,
		void (*exception_request)(), int (*command_request)())
{
  int	 nfds;
  fd_set socks;
  int	 stat;
  struct timeval time_out;

  //time_out.tv_sec = (int)floor(timeout / 1000);
  time_out.tv_sec = timeout / 1000;
  time_out.tv_usec = (timeout - time_out.tv_sec * 1000) * 1000 ;

  FD_ZERO(&socks);
  FD_SET(client_port, &socks);

  nfds = client_port+1;
  stat=select(nfds, &socks, 0, 0 /*&exceptions*/, &time_out);

  if (stat>0) {
    if (FD_ISSET(client_port, &socks)) { 
      if(command_request){
        if((*command_request)(client_port) == -1){
          close(client_port);
	  return 0;
        }
      }
      FD_CLR(client_port, &socks);
    }
  }

  return 1; 
}
#endif /* no use */


////////////// TEST Implementation
//

void init_SockProfile(){
  int i;

  SOCKET_LOCK();

  for(i=0;i<FD_SETSIZE;i++){
    memset(&SockProfile[i], 0, sizeof(struct sockport_profile));
  }

  SOCKET_UNLOCK();

  return;
}

void clear_SockProfile(int fd){

  SOCKET_LOCK();

  memset(&SockProfile[fd], 0, sizeof(struct sockport_profile));

  SOCKET_UNLOCK();
  
  return;
}

/*
void set_SockProfile_type(int fd, int type){

  SOCKET_LOCK();
  
  SockProfile[fd].type = type;

  SOCKET_UNLOCK();
  return;
}
*/
void set_SockProfile_arg(int fd, void *arg){
  
  SOCKET_LOCK();
  
  SockProfile[fd].arg = arg;

  SOCKET_UNLOCK();
  return;
}

void *get_SockProfile_arg(int fd)
{
  return SockProfile[fd].arg;
}


/*
#ifdef USE_SHMC
void set_SockProfile_shmc(int fd, Shmc *shmc)
{
  SOCKET_LOCK();
  
  SockProfile[fd].shmc = shmc;

  SOCKET_UNLOCK();
}
#endif
*/
/*
void set_SockProfile_service(int fd,
		int (*command_request)()){

  SOCKET_LOCK();
  
  SockProfile[fd].command_proc = command_request;

  SOCKET_UNLOCK();

  return;
}
*/

static void get_Servers_from_SockProfile(int *res, int *n, int start_fd){
  int i;
  int current;

  SOCKET_LOCK();

  for(i=start_fd, current=0;i<FD_SETSIZE;i++){
    if(SockProfile[i].type == SOCK_SERVER){
      res[ current++ ]=i;
    }
  }
  *n=current;

  SOCKET_UNLOCK();

  return;
}

static void copy_sockport_profile(int dist, int src){

  SOCKET_LOCK();
  
  memcpy(&SockProfile[dist], &SockProfile[src], sizeof(struct sockport_profile));

  SOCKET_UNLOCK();

  return;
}

int set_SockProfile(int fd, int type,
		int (*connection_request)(),
		int (*disconnect_request)(),
		int (*command_request)()){
  SOCKET_LOCK();
  
  SockProfile[fd].type = type;
  SockProfile[fd].command_proc = command_request;
  SockProfile[fd].connection_proc = connection_request;
  SockProfile[fd].disconnect_proc = disconnect_request;

  FD_SET(fd, &main_socket_bits);

  SOCKET_UNLOCK();

  return fd;
}

int clear_socket_profile(int fd, void (*clear_func)()){
  clear_SockProfile(fd);

  SOCKET_LOCK();

  FD_CLR(fd, &main_socket_bits);

  SOCKET_UNLOCK();

  if(clear_func) (*clear_func)(fd);

  return fd;
}

static int check_current_sockbits(int fd){
  if(FD_ISSET(fd, &main_socket_bits)) return 1;

  return 0;
}

static int next_active_socket(fd_set *socks, int start_fd){
  int i;

  for (i = start_fd; i < FD_SETSIZE; i++)
    if (FD_ISSET(i, socks))  return i;

  return -1;
}

int select_socket_ports( fd_set *sockbits, struct timeval time_out)
{
  int	 nfds;
  fd_set socks;
  int	 stat, res, newsock;
  int active_port;
  int (*func)();
  GIOP_ConnectionHandler h;
//  struct timeval tv;

  FD_ZERO(&socks);
  nfds = copy_fds_set(&socks, sockbits);

//  tv.tv_sec = time_out->tv_sec;
//  tv.tv_usec = time_out->tv_usec;
  res = stat = select(nfds, &socks, 0, 0, &time_out);
  active_port = 0;

  while( stat > 0 ){
    if((active_port = next_active_socket(&socks, active_port) ) < 0 ){
	  ORB_ERROR_OUT1("Invalid port is active...\n");
      return -1;
    }

    SOCKET_LOCK();

    func = SockProfile[active_port].command_proc;

    SOCKET_UNLOCK();

    switch (SockProfile[active_port].type){
      case SOCK_SERVER:
#if DEBUG
	ORB_ERROR_OUT2("Request Connection fd=%d\n", active_port);
#endif
        newsock = accept_connection(active_port);     /* new connection */

        if (newsock>0) {
          FD_SET(newsock, sockbits);
	  copy_sockport_profile(newsock, active_port);

	  SOCKET_LOCK();

          SockProfile[newsock].type = SOCK_SERVICE;

	  SOCKET_UNLOCK();

#if DEBUG
	  ORB_ERROR_OUT2("New Connection fd=%d\n", newsock);
#endif
	}

        func = SockProfile[active_port].connection_proc;
	if(func) (*func)(active_port, SockProfile[active_port].arg );

        FD_CLR(active_port, &socks);
        stat--;
	break;
      case SOCK_CLIENT:
      case SOCK_SERVICE:
	h.type = CONNECTION_TYPE_SOCKET;
	h.sock = active_port;

        if(func &&  (*func)(&h) == -1 ){
          func = SockProfile[active_port].disconnect_proc;
	  if(func) (*func)(active_port, SockProfile[active_port].arg );

#if DEBUG
	  ORB_ERROR_OUT2("Request Close Connection fd=%d\n", active_port);
#endif
	  close(active_port);

          FD_CLR(active_port, sockbits);

	  SOCKET_LOCK();

          SockProfile[active_port].type = SOCK_CLOSED;

	  SOCKET_UNLOCK();
	}

        FD_CLR(active_port, &socks);
        stat--;
	break;
      default:
		ORB_ERROR_OUT1("Socket already closed...\n");
        stat--;
	break;
      }
  }

  return res; 
}

void init_socket_servers()
{
  int i;
  int sockServers[FD_SETSIZE];
  int nSock;

  get_Servers_from_SockProfile(sockServers, &nSock, 0);

  SOCKET_LOCK();

  FD_ZERO(&main_socket_bits);
  for(i=0;i<nSock;i++){ FD_SET(sockServers[i], &main_socket_bits); }

  SOCKET_UNLOCK();
}

void select_socket_servers(struct timeval time_out)
{
  select_socket_ports( &main_socket_bits, time_out);
}

