// -*- C++ -*-
/*!
 * @file  Condition.h
 * @brief Condition variable for POSIX
 * @date  $Date$
 * @author Noriaki Ando <n-ando@aist.go.jp>
 *
 * Copyright (C) 2008
 *     Noriaki Ando
 *     Task-intelligence Research Group,
 *     Intelligent Systems Research Institute,
 *     National Institute of
 *         Advanced Industrial Science and Technology (AIST), Japan
 *     All rights reserved.
 *
 * $Id$
 *
 */

#ifndef COIL_CONDITION_H
#define COIL_CONDITION_H

#include <kernel.h>
#include <t_syslog.h>
#include <coil/Mutex.h>

namespace coil
{
  typedef struct
  {
    // Semaphore used to queue up task waiting for the condition to
    // become signaled.
    ID semid;
  } pthread_cond_t;

  inline int pthread_cond_init(coil::pthread_cond_t *cv, ID semid)
  {
	cv->semid = semid;
	return 0;
  }

  inline ER pthread_cond_wait(coil::pthread_cond_t *cv, coil::Mutex *mutex)
  {
    ID ercd;

    mutex->unlock();
    ercd = wai_sem(cv->semid);

    return ercd;
  }

  inline ER pthread_cond_signal(pthread_cond_t *cv)
  {
	return sig_sem(cv->semid);
  }

  inline ER pthread_cond_broadcast(pthread_cond_t *cv)
  {
	return ini_sem(cv->semid);
  }

  inline ER pthread_cond_timedwait(pthread_cond_t *cv, coil::Mutex *mutex,
		  TMO tmout)
  {
    return twai_sem(cv->semid, tmout);
  }

  template <class M>
  class Condition
  {
  public:
    Condition(M& mutex)
      : m_mutex(mutex)
    {
    }

    ~Condition()
    {
    }

    inline void init(ID semid)
    {
      coil::pthread_cond_init(&m_cond, semid);
    }

    inline void signal()
    {
      m_mutex.trylock();
      coil::pthread_cond_signal(&m_cond);
    }

    inline void broadcast()
    {
      m_mutex.trylock();
      coil::pthread_cond_broadcast(&m_cond);
    }

    ER wait()
    {
      m_mutex.trylock();
      return coil::pthread_cond_wait(&m_cond, &m_mutex);
    }

    ER wait(long second, long nano_second = 0)
    {
      TMO tmout = second * 1000 + nano_second / 1000000;
      m_mutex.trylock();
      return pthread_cond_timedwait(&m_cond, &m_mutex, tmout);
    }

  private:
	Condition(const Mutex&);
	Condition& operator=(const Mutex &);
    coil::pthread_cond_t m_cond;
    M& m_mutex;
  };
};
#endif // COIL_CONDITION_H
