// -*- C++ -*-
/*!
 * @file  UUID.h
 * @brief RT-Middleware Service interface
 * @date  $Date$
 * @author Noriaki Ando <n-ando@aist.go.jp>
 *
 * Copyright (C) 2008
 *     Noriaki Ando
 *     Task-intelligence Research Group,
 *     Intelligent Systems Research Institute,
 *     National Institute of
 *         Advanced Industrial Science and Technology (AIST), Japan
 *     All rights reserved.
 *
 * $Id$
 *
 */

#ifndef COIL_UUID_H
#define COIL_UUID_H

#include <itron.h>
#include <kernel.h>
#include <coil/config_coil.h>
#include <coil/Time.h>
#include <coil/String.h>
#include <string.h>
#include <stdio.h>
#include "ethmac.h"
#include "stringlib.h"


typedef struct {
  uint32_t  time_low;
  uint16_t  time_mid;
  uint16_t  time_hi_and_version;
  uint8_t   clock_seq_hi_and_reserved;
  uint8_t   clock_seq_low;
  uint8_t   node[6];
} uuid_t;

namespace coil
{
  typedef SYSTIM uuid_time_t;

  typedef struct {
    int32_t  tv_sec;
    int32_t  tv_usec;
    uint16_t  cs;       /* saved clock sequence */
  } uuid_state;

/* set the following to the number of 100ns ticks of the actual
   resolution of your system's clock */
#define UUIDS_PER_TICK 1024

  static uuid_state uuid_st;

  inline void read_macaddr(uint8_t mac[6])
  {
     memcpy(mac, get_ifaddr(), 6);
  }

  /* generate UUID version 1: time, clock and node based */
  inline void uuid_generate(uuid_t *uuid)
  {
    struct timeval tv;
    uint64_t ts;
    uint8_t mac[6];
    uint16_t clock_seq = coil::uuid_st.cs;

    /* determine current system time and sequence counter */
    /* determine current system time */
    coil::gettimeofday(&tv, 0);

    /* get MAC address */
    read_macaddr(mac);

    /* check whether system time changed since last retrieve */
    if (!(tv.tv_sec  == coil::uuid_st.tv_sec
        && tv.tv_usec == coil::uuid_st.tv_usec)) {
      clock_seq = 0;
    } else {
      clock_seq++;
    }

    /* convert from (sec,usec) to 100*nsec format */
    ts = tv.tv_sec * 1000000 + tv.tv_usec / 10;

    /* store back new system time */
    coil::uuid_st.tv_sec = tv.tv_sec;
    coil::uuid_st.tv_usec = tv.tv_usec;
    coil::uuid_st.cs = clock_seq;

    /* stuff fields into the UUID */
    uuid->time_low = (uint32_t)(ts & 0xFFFFFFFF);
    uuid->time_mid = (uint16_t)((ts >> 32) & 0xFFFF);
    uuid->time_hi_and_version = (uint16_t)((ts >> 48) & 0x0FFF);
    uuid->time_hi_and_version |= (1 << 12);
    uuid->clock_seq_low = clock_seq & 0xFF;
    uuid->clock_seq_hi_and_reserved = (clock_seq & 0x3F00) >> 8;
    uuid->clock_seq_hi_and_reserved |= 0x80;
    /* use real regular MAC address */
    memcpy(uuid->node, mac, 6);
  }

  inline void uuid_clear(uuid_t *uuid)
  {
    uuid->time_low = 0;
    uuid->time_mid = 0;
    uuid->clock_seq_low = 0;
    uuid->clock_seq_hi_and_reserved = 0;
    uuid->time_hi_and_version = 0;
    memset(uuid->node, 0, 6);
  }

  inline void uuid_unparse(uuid_t *uuid, char* buf)
  {
    int ws = t_vsnprintf(buf, 32, "%8.8x-%4.4x-%4.4x-%2.2x%2.2x-",
            uuid->time_low, uuid->time_mid, uuid->time_hi_and_version,
            uuid->clock_seq_hi_and_reserved, uuid->clock_seq_low);
    for (int i = 0; i < 6; i++) {
        ws += t_vsnprintf(&buf[ws], 32, "%2.2x", uuid->node[i]);
    }
  }

  class UUID
  {
    uuid_t _uuid;
    char buf[37];
  public:
    UUID();
    UUID(uuid_t*);
    const char* to_string();
  };

  class UUID_Generator
  {
  public:
    UUID_Generator();

    void init();
    UUID* generateUUID(int n, int h);
  };
};
#endif // COIL_UUID_H
